import logging

from dbnd import log_metrics
from py4j.protocol import Py4JJavaError
from pydeequ.repository import MetricsRepository


logger = logging.getLogger(__name__)


class DeequMetricsRepository(object):
    def save(self, result_key, analyzer_context):
        pass

    def loadByKey(self, result_key):
        pass

    def load(self):
        pass

    class Java:
        implements = ["com.amazon.deequ.repository.MetricsRepository"]


class DbndDeequMetricsRepository(DeequMetricsRepository):
    def __init__(self, spark_session):
        self.jvm = spark_session._jvm
        try:
            # check if dbnd jar was loaded
            self.jvm.ai.databand.deequ.NoopMetricsRepository()
            self.java_libs_loaded = True
        except Py4JJavaError:
            logger.error(
                "Unable to load Databand Java SDK. Perhaps you forgot to add dbnd to spark classpath?"
            )
            self.java_libs_loaded = False

    def save(self, result_key, analyzer_context):
        if not self.java_libs_loaded:
            logger.warn(
                "Databand Java SDK was not loaded. Deequ metrics won't be logged"
            )
            return
        df_name = (
            self.jvm.scala.collection.JavaConverters.mapAsJavaMapConverter(
                result_key.tags()
            )
            .asJava()
            .getOrDefault("name", "data")
        )
        converter = self.jvm.ai.databand.deequ.DeequToDbnd(df_name, analyzer_context)
        metrics = dict(converter.metrics())
        log_metrics(metrics)

    def loadByKey(self, result_key):
        return self.jvm.scala.Option.empty()


class DbndMetricsRepository(MetricsRepository):
    def __init__(self, spark_session):
        self.repository = DbndDeequMetricsRepository(spark_session)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json
import re
from json import JSONDecodeError
from urllib.parse import urlparse, parse_qsl


def unquote_to_bytes(string):
    """unquote_to_bytes('abc%20def') -> b'abc def'."""
    # Note: strings are encoded as UTF-8. This is only an issue if it contains
    # unescaped non-ASCII characters, which URIs should not.
    if not string:
        # Is it a string-like object?
        string.split
        return b''
    if isinstance(string, str):
        string = string.encode('utf-8')
    bits = string.split(b'%')
    if len(bits) == 1:
        return string
    res = [bits[0]]
    append = res.append
    # Delay the initialization of the table to not waste memory
    # if the function is never called
    global _hextobyte
    if _hextobyte is None:
        _hextobyte = {(a + b).encode(): bytes([int(a + b, 16)])
                      for a in _hexdig for b in _hexdig}
    for item in bits[1:]:
        try:
            append(_hextobyte[item[:2]])
            append(item[2:])
        except KeyError:
            append(b'%')
            append(item)
    return b''.join(res)


_asciire = re.compile('([\x00-\x7f]+)')


def unquote(string, encoding='utf-8', errors='replace'):
    """Replace %xx escapes by their single-character equivalent. The optional
    encoding and errors parameters specify how to decode percent-encoded
    sequences into Unicode characters, as accepted by the bytes.decode()
    method.
    By default, percent-encoded sequences are decoded with UTF-8, and invalid
    sequences are replaced by a placeholder character.

    unquote('abc%20def') -> 'abc def'.
    """
    if '%' not in string:
        string.split
        return string
    if encoding is None:
        encoding = 'utf-8'
    if errors is None:
        errors = 'replace'
    bits = _asciire.split(string)
    res = [bits[0]]
    append = res.append
    for i in range(1, len(bits), 2):
        append(unquote_to_bytes(bits[i]).decode(encoding, errors))
        append(bits[i + 1])
    return ''.join(res)


# Python automatically converts all letters to lowercase in hostname
# See: https://issues.apache.org/jira/browse/AIRFLOW-3615
def _parse_netloc_to_hostname(uri_parts):
    """Parse a URI string to get correct Hostname."""
    hostname = unquote(uri_parts.hostname or '')
    if '/' in hostname:
        hostname = uri_parts.netloc
        if "@" in hostname:
            hostname = hostname.rsplit("@", 1)[1]
        if ":" in hostname:
            hostname = hostname.split(":", 1)[0]
        hostname = unquote(hostname)
    return hostname


class Connection:
    def __init__(self, uri):
        self.conn_type = None
        self.host = None
        self.schema = None
        self.login = None
        self.password = None
        self.port = None
        self.extra = None
        self._parse_from_uri(uri)

    @property
    def extra_dejson(self):
        """Returns the extra property by deserializing json."""
        obj = {}
        if self.extra:
            try:
                obj = json.loads(self.extra)
            except JSONDecodeError as e:
                pass
        return obj

    def _parse_from_uri(self, uri):
        uri_parts = urlparse(uri)
        conn_type = uri_parts.scheme
        if conn_type == 'postgresql':
            conn_type = 'postgres'
        elif '-' in conn_type:
            conn_type = conn_type.replace('-', '_')
        self.conn_type = conn_type
        self.host = _parse_netloc_to_hostname(uri_parts)
        quoted_schema = uri_parts.path[1:]
        self.schema = unquote(quoted_schema) if quoted_schema else quoted_schema
        self.login = unquote(uri_parts.username) \
            if uri_parts.username else uri_parts.username
        self.password = unquote(uri_parts.password) \
            if uri_parts.password else uri_parts.password
        self.port = uri_parts.port
        if uri_parts.query:
            self.extra = json.dumps(dict(parse_qsl(uri_parts.query, keep_blank_values=True)))



from dbnd import parameter
from dbnd._core.utils.http.constants import NO_AUTH
from dbnd_spark.spark_config import SparkEngineConfig


class LivySparkConfig(SparkEngineConfig):
    # we don't want spark class to inherit from this one, as it should has Config behaviour
    _conf__task_family = "livy"
    _conf__help_description = "livy configuration"

    url = parameter(description="livy connection url (for example: http://livy:8998)")[
        str
    ]

    auth = parameter(
        description="livy auth , support list are None, Kerberos, Basic_Access",
        default=NO_AUTH,
    )[str]

    user = parameter(description="livy auth , user", default="")[str]

    password = parameter(description="livy auth , password", default="")[str]

    ignore_ssl_errors = parameter(description="ignore ssl error", default=False)[bool]

    post_submit_hook = parameter(
        description="User code to run after livy batch submit (a reference to a function)"
        "expected interface (LivySparkCtrl, Dict[str, Any]) -> None",
        default=None,
    )[str]

    retry_on_status_error = parameter(
        description="retries http requests if status code is not accepted", default=0
    )[int]

    def get_spark_ctrl(self, task_run):
        from dbnd_spark.livy.livy_spark import LivySparkCtrl

        return LivySparkCtrl(task_run)



import logging

import dbnd

from dbnd import dbnd_config, register_config_cls
from dbnd._core.configuration.config_readers import get_environ_config_from_dict
from dbnd._core.configuration.environ_config import (
    _debug_init_print,
    spark_tracking_enabled,
)
from dbnd._core.settings.histogram import HistogramConfig
from dbnd_spark.livy.livy_spark_config import LivySparkConfig
from dbnd_spark.spark_session import has_pyspark_imported


logger = logging.getLogger(__name__)


@dbnd.hookimpl
def dbnd_setup_plugin():
    from dbnd_spark.local.local_spark_config import SparkLocalEngineConfig
    from dbnd_spark.spark_bootstrap import dbnd_spark_bootstrap

    register_config_cls(SparkLocalEngineConfig)
    register_config_cls(LivySparkConfig)

    dbnd_spark_bootstrap()

    if has_pyspark_imported() and spark_tracking_enabled():
        config_store = read_spark_environ_config()
        dbnd_config.set_values(config_store, "system")
    else:
        _debug_init_print(
            "spark conf is not loaded since pyspark is not imported or DBND__ENABLE__SPARK_CONTEXT_ENV is not set"
        )


def read_spark_environ_config():
    logger.debug("running read_spark_environ_config")

    from pyspark import SparkContext

    _debug_init_print("creating spark context to read spark conf")
    spark_conf = SparkContext.getOrCreate().getConf()
    spark_conf = dict(spark_conf.getAll())
    spark_conf = {key.lstrip("spark.env."): value for key, value in spark_conf.items()}

    return get_environ_config_from_dict(spark_conf, "environ")
